#include "qcl.h"

extern int		mdm;
extern int		capture_fd, capture_on;
extern unsigned	kbd_mode, okbd_mode;
extern int		yyline;
extern struct	state_name state_table[ MAX_STATE ];
extern char		input_buf[];
extern char		*line_table[], *line_ptr;
extern char		debug, echo, secure, quiet;
extern int		call_stack[], call_level, pass;
extern struct	var_entry *retval, *cmdval, *collect, *failed, *switcher;

/*----------------------------------------------------------------------------
IMPORT:	src - Pointer to string to work on.
		num - Number of chars to take from front (+) or end (-) of string.
EXPORT: Pointer to newly allocated string containing result of operation.
----------------------------------------------------------------------------*/
char *
Take( char *src, int num ) {
	char 	*p, *p1;
	int 	len = strlen( src );
	static char *temp = NULL;

	if ( temp ) _Free( temp );
	temp = NULL;
	if ( num < 0 ) {						/* Take from end of string		*/
		num = -num;
		if ( num > len ) num = len;
		p = &src[ len - num ];
		}
	else {
		if ( num > len ) num = len;			/* Take from start of string	*/
		p = src;
		}
	memcpy( temp = p1 = _Malloc( num + 1 ), p, num );
	p1[ num ] = '\0';
	return p1;
	}

/*----------------------------------------------------------------------------
IMPORT:	src - Pointer to string to work on.
		set - Pointer to string containing character set.
		dir = 1: Work from start of string.
		dir = 0: Work from end of string.
EXPORT: Pointer to newly allocated string containing result of operation.
----------------------------------------------------------------------------*/
char *
Take_str( char *src, char *s, int dir ) {
	char 	*p;
	int 	pos;
	static char *temp = NULL;

	if ( temp ) _Free( temp );
	temp = NULL;
	if ( !*src  ||  !*s ) return src;	// Handle null strings
	if ( dir ) {							/* From start of string			*/
		if ( *s == '^' ) pos = strspn( src, ++s );/* Take chars while in set*/
		else			 pos = strcspn( src, s );/* Take chars while !in set*/
		memcpy( temp = p = _Malloc( pos+1 ), src, pos );
		p[pos] = '\0';
		return p;
		}
	if ( *s == '^' )	pos = rstrspn( src, ++s );
	else				pos = rstrcspn( src, s );
	return temp = mkstr( &src[pos] );
	}

rstrspn( char *str1, char *set2 ) {
	char 	*s, *set;
	int 	len = strlen( str1 ) - 1;

	for( s = str1 + len; s != str1; --s ) {
		for( set = set2; *set; ++set )
			if ( *s == *set ) goto more;
		break;
more:;
		}
	return s - str1 + 1;
	}

rstrcspn( char *str1, char *set2 ) {
	char	*s, *set;
	int 	len = strlen( str1 );

	for( s = str1 + len; s != str1; --s )
		for( set = set2; *set; ++set )
			if ( *s == *set ) goto done;
done:
	return s != str1 ? s - str1 + 1 : 0;
	}

/*----------------------------------------------------------------------------
IMPORT:	src - Pointer to string to work on.
		num - Number of strings to take from front (+) or end (-) of string.
EXPORT: Pointer to newly allocated string containing result of operation.
----------------------------------------------------------------------------*/
char *
Drop( char *src, int num ) {
	char	*p;
	int		len = strlen( src );
	static char *temp = NULL;

	if ( temp ) _Free( temp );
	temp = NULL;
	if ( num < 0 ) {						/* Drop from end of string		*/
		num *= -1;
		if ( num > len ) num = len;
		strncpy( temp = p = _Malloc( len - num + 1 ), src, len - num );
		p[ len - num ] = '\0';
		return p;
		}
	if ( num > len ) num = len;				/* Drop from start of string	*/
	return temp = mkstr( &src[ num ] );
	}

/*----------------------------------------------------------------------------
IMPORT:	src - Pointer to string to work on.
		set - Pointer to string containing character set.
		dir = 1: Work from start of string.
		dir = 0: Work from end of string.
EXPORT: Pointer to newly allocated string containing result of operation.
----------------------------------------------------------------------------*/
char *
Drop_str( char *src, char *set, int dir ) {
	char	*p, *s = set;
	int		pos;
	static char *temp = NULL;

	if ( temp ) _Free( temp );
	temp = NULL;
	if ( !*src  ||  !*s ) return src;	/* Handle null strings			*/
	if ( dir ) {						/* From start of string				*/
		if ( *s == '^' )				/* Drop chars while in set			*/
			pos = strspn( src, ++s );
		else pos = strcspn( src, s );	/* Else take chars while not in set	*/
		return temp = mkstr( &src[ pos ] );
		}
	if ( *s == '^' )					/* From end of string				*/
		pos = rstrspn( src, ++s );		/* While in set						*/
	else pos = rstrcspn( src, s );		/* Else take chars while not in set	*/
	memcpy( temp = p = _Malloc( pos + 1 ), src, pos + 1 );
	p[ pos ] = '\0';
	return p;
	}

char *
Tolower( char *str ) {
	char	*s = str, *d, *p;

	p = d = (char *) _Malloc( strlen( s ) + 1 );
	while( *s )	{
		*d++ = tolower( *s );
		++s;
		}
	*d = '\0';
	return p;
	}

char *
Toupper( char *s ) {
	unsigned char	*d, *p;

	p = d = (char *) _Malloc( strlen( s ) + 1 );
	while( *s ) {
		*d++ = toupper( *s );
		++s;
		}
	*d = '\0';
	return p;
	}

void
Emit( char *str, int mdm_echo ) {
	write_mdm( 0, mdm_echo, str );
	}

void
Collect( unsigned t1, unsigned t2 ) {
	int		nchrs;

	switcher = collect;			/* Revert back to collect$ for cases.	*/
	input_buf[0] = '\0';

	abort_check(0);
	if ( (nchrs = dev_read( mdm, input_buf, BUFFER_SIZE, BUFFER_SIZE, t2, 10*t1, 0, 0 )) == -1 ) {
		perror( "dev_read" );
		return;
		}
	if ( capture_fd != -1 && capture_on ) write( capture_fd, input_buf, nchrs );
	abort_check(0);
	nchrs = strip_nulls( &input_buf, nchrs );
	qterm_fput( &input_buf, nchrs );
	}

Waitfor( char *pattern, unsigned time_limit, unsigned quiet_time ) {
	int		nchrs;

	input_buf[0] = '\0';

	abort_check(0);
	if ( (nchrs = dev_read( mdm, input_buf, BUFFER_SIZE, BUFFER_SIZE, quiet_time, 10*time_limit, 0, 0 )) == -1 ) {
		perror( "dev_read" );
		return 0;
		}
	if ( capture_fd != -1 && capture_on ) write( capture_fd, input_buf, nchrs );
	abort_check(0);
	nchrs = strip_nulls( &input_buf, nchrs );
	qterm_fput( &input_buf, nchrs );
	return fnmatch( pattern, input_buf, 0 ) == 0;
	}

Case( char *pattern ) {
	if ( !switcher->vstr ) {
		yyerror( "Bad string for case statement." );
		done( -1 );
		}
	return fnmatch( pattern, switcher->vstr, 0 ) == 0;
	}

void
enter( char *state_name, int is_goto ) {
	extern int	last_state;
	int			i;

	for ( i = 0; i < last_state; ++i )
		if ( !strcmp( state_table[i].name, state_name ) ) {
			if ( is_goto )	Goto( i );
			else			Call( i );
			return;
			}
	yyerror( "Unknown state name '%s'", state_name );
	done( -1 );
	}

void
Goto( int state ) {
	if ( state < 0  ||  state > MAX_STATE )
		yyerror( "Goto to invalid state %d\r\l", state );

	if ( !state_table[state].first_line )
		yyerror( "Transfer to non-existant state %d\r\l", state );

	yyline = state_table[ state ].first_line;
	line_ptr = line_table[ yyline ];
	if ( pass == 2  &&  debug )
		fprintf( stderr, "\r\lEXEC %d. %s\l", yyline, line_ptr );
	}

void
Call( int state ) {
	if ( call_level >= 100 )	yyerror( "Call stack overflow.\n" );

	call_stack[ call_level++ ] = yyline + 1;
	Goto( state );
	}

void
Return( int val ) {
	if ( call_level <= 0 )		/* If highest level, then do an exit		*/
		done( val );
	retval->vint = val;
	yyline = call_stack[ --call_level ];
	line_ptr = line_table[ yyline ];
	if ( pass == 2  &&  debug )
		fprintf( stderr, "\r\lEXEC %d. %s\l", yyline, line_ptr );
	}

compare( int val1, int operator, unsigned val2 ) {
	switch( operator ) {
		case '==':	return val1 == val2;
		case '!=':	return val1 != val2;
		case '<=':	return val1 <= val2;
		case '>=':	return val1 >= val2;
		case '>' :	return val1 >  val2;
		case '<' :	return val1 <  val2;
		default:	fprintf( stderr, "qcl:  Bad comparison operator!\r\l" );
					return -1;
		}
	}

compute( int val1, int operator, unsigned val2 ) {
	switch( operator ) {
		case '*':	return val1 * val2;
		case '/':	return val1 / (val2 ? val2 : 1);
		case '%':	return val1 % (val2 ? val2 : 1);
		default:	fprintf( stderr, "qcl:  Bad compute operation!\r\l" );
					return -1;
		}
	}

void
Jump( int relative, int line ) {
	if ( relative )	line = yyline + relative * line;

	if ( line < 0  ||  line > 1000 )
		yyerror( "Goto to invalid line %d.\n", line );

	line_ptr = line_table[yyline = line];
	if ( pass == 2  &&  debug )
		fprintf( stderr, "\r\lEXEC %d. %s\l", yyline, line_ptr );
	}

int
Display( int coord, char *string, unsigned attr ) {
	char	*str = string, *p = str;
	int		r, c;
   	
	if ( quiet ) return 0;

	if ( coord == -1 ) {
		r = term_state.row;
		c = term_state.col;
		}
	else {
		r = (coord >> 8);
		c = (coord & 0xff);
		}

	while( *p ) {
		for( ; *p != '\r' && *p != '\l' && *p; p++ );
		if ( p - str ) term_type( r, c, str, p - str, attr );
		r = term_state.row;
		c = term_state.col;
		for( ; (*p == '\r' || *p == '\l') && *p; p++ ) {
			switch( *p ) {
				case '\r':	c = 0;
							break;
				case '\l':	if ( ++r >= term_state.num_rows ) {
								r--;
								term_delete_line( 0, 1 );
								}
							break;
				}
			}
		str = p;
		}
	term_cur( r, c );
	term_flush();
	return strlen( string );
	}

char *
Input() {
	char		buf[ 82 ];

	buf[0] = '\0';
	if ( isatty(0) && !quiet ) {
		a_dev_mode(0, _DEV_MODES, _DEV_MODES);
		fgetln( stdin, buf, 80 );
		a_dev_mode(0, kbd_mode, _DEV_MODES);
		if ( strlen( buf ) > 1 )
			buf[strlen(buf)-1] = '\0';
		}
	return mkstr( buf );
	}

#define	ETX		0x03
#define	ACK		0x06
#define	NAK		0x15
#define	CAN		0x18

/* For this to be successful Qterm must have been started with type 0 protocommands
   enabled...
*/
proto_emit( int retry_limit, int wait, char *cmd ) {
	int					len, count = 0;
	uchar				sum, ack;
	char				cs[5], buf[90], *p;

	if ( ( len = strlen( cmd ) ) >= 80 )	return -1;
	sprintf( cs, "%02d", len );
	sum = cs[0] + cs[1];		/* Checksum includes byte count and text	*/
	p = cmd;
	while( *p ) sum += *p++;

	/*	Emit the protocommand, try to read 1 character, if 'wait' seconds pass
		and we get nothing from the other end loop back to the beginning.
		Only do this for 10 seconds at most (retry_limit loops)
	*/
	do {
		ack = NAK;
		sprintf( buf, "\033^0%s%s%03d\r", cs, cmd, sum );
		if ( write( mdm, buf, strlen(buf) ) != strlen( buf ) )
			fprintf( stderr, "qcl: Unable to output proto_command\r\l" );

		if ( dev_read( mdm, &ack, 1, 1, wait*10, 0, 0, 0 ) == -1 )
			ack = ETX;
		} while( ack == NAK   &&  ++count <= retry_limit );

	if ( ack == ACK )					return 1;
	if ( ack == NAK  ||  ack == CAN )	return 0;
	return -1;
	}

void
abort_check( int sig ) {
	int			c = 0;
	extern char qterm;

	if ( (quiet || !dev_ischars(0)) && !sig || secure ) return;

	/* Need ' ', Enter or escape for abort prompt	*/
	if ( sig != SIGINT ) c = getc( stdin );
	if ( sig == SIGINT  ||  c == ' '  ||  c == 27  ||  c == '\r' ) {
		write( 2, "\r\lQCL Script Aborted\r\l", 22 );
		done( -1 );
		}
	}
