/*				 9-Sep-88  1:02:09pm										*/

/*--------------------------------------------------------------------------*/
/*  History: cond.c, V0.0, ??-Nov-86 11:24:14am, Dan Dodge,	Baseline		*/
/*  				 V0.1, 13-Jan-87 11:24:14am, Dan Hildebrand				*/
/*----------------------------------------------------------------------------

     This module  converts  structured  statements  into  conventional  if
     statements with the appropriate jumps inserted.

----------------------------------------------------------------------------*/
#include "qcl.h"
#include "y.tab.h"

extern int yychar, yyline, last_line;
extern char *line_table[], *line_ptr;
extern char *line_ptr_save;

struct	control_entry {
			int control_type;
			int control_line;
			} control_stack[ 20 ];

int		control_level, block_end_line;

char *
skip_kword( char *buf ) {
	char *p;

	for( p = buf; *p == ' '  ||  *p == '\t'; ++p ) ;	/* Skip leading white */
	for( ; isalpha(*p); ++p ) ;							/* Skip keyword		*/
	return( p );
	}

void
beg_cond( int type ) {
	struct control_entry *cp;
	char *tname;
	char buf[100];

	if ( ++control_level >= 20 ) {
		yyerror( "Control nesting too deep." );
		done( -1 );
		}

	if ( yychar == -1 )	yychar = yylex();

	if ( type == _REPEAT ) {	/* Must make up an assignment and if !var-- */
		memmove( &line_table[yyline+1],
				 &line_table[yyline],
				 (last_line-yyline+1)*sizeof(char *) );
		last_line += 1;

		sprintf( buf, "%s=%.*s\n", tname = make_unique_name()->vname,
					line_ptr_save - skip_kword( line_table[yyline] ),
					skip_kword( line_table[yyline] ) );
		line_table[ yyline ] = mkstr( buf );

		sprintf( buf, "if %s--", tname );
		line_table[ yyline + 1 ] = mkstr( buf );

		++yyline;
		type = _WHILE;
		}

	cp = &control_stack[ control_level ];
	cp->control_type = type;
	cp->control_line = yyline;

	/*
	 * We always turn
	 *		if cond statement
	 * into
	 *		if cond jump line-num
	 *			statement
	 */

	if ( yychar != '{' ) {
		memmove( &line_table[yyline+3],
				 &line_table[yyline+1],
				 (last_line-yyline)*sizeof(char *) );
		last_line += 2;
		line_table[ yyline + 1 ] = mkstr( line_ptr_save );
		line_table[ yyline + 2 ] = "}\n";
		}
	*line_ptr_save = '\0';
	line_ptr = "{\n";
	yychar = -1;
	}

void
else_cond() {
	struct control_entry *cp;

	cp = &control_stack[ control_level ];
	if ( cp->control_type != _IF )		yyerror( "Misplaced ELSE." );

	end_cond( 1 );
	beg_cond( _ELSE );
	line_table[yyline] = "";
	}

char *
not_cond( char *buf ) {
	char *p;

	p = skip_kword( buf );

	memmove( p + 1, p, strlen(p) + 1 );
	*p = '!';

	return( p );
	}

void
end_cond( int inc ) {
	struct control_entry *cp;
	char *p;
	char buf[100];

	if ( control_level < 1 ) {
		yyerror( "Too many closing braces '}'." );
		done( -1 );
		}

	cp = &control_stack[ control_level-- ];
	sprintf( buf, "%s jump %d\n",
					line_table[cp->control_line], block_end_line + inc);
	if ( cp->control_type == _IF		||
		 cp->control_type == _WHILE		||
		 cp->control_type == _REPEAT		)	p = not_cond( buf );

	if ( cp->control_type == _WHILE ) /* Turn into an if */
		sprintf( buf, "if%s", p );
	line_table[ cp->control_line ] = mkstr( buf );

	if ( cp->control_type == _WHILE ) {
		sprintf( buf, "jump %d\n", cp->control_line );
		line_table[ block_end_line ] = mkstr( buf );
		}
	else if ( cp->control_type == _REPEAT ) {
		sprintf( buf, "loop T1,%d\n", cp->control_line );
		line_table[ block_end_line ] = mkstr( buf );
		}
	else {
		memmove( &line_table[ block_end_line ],
				 &line_table[ block_end_line+1 ],
				 ( last_line-block_end_line+1 ) * sizeof( char * ) );
		--last_line;
		--yyline;
		}
	}

void
split_line( int line ) {
	line_table[ line ] = skip_kword( line_table[ line ] );
	memmove( &line_table[ line+1 ],
			 &line_table[ line ],
			 ( last_line - line ) * sizeof( char * ) );
	line_table[ line ] = "\n";			/* Will be filled in by caller	*/
	}
