#include "qcl.h"
#include <ctype.h>

int		mdm;
unsigned	kbd_mode, okbd_mode, omdm_mode;
int		capture_fd;
FILE	*nest_fp[NEST_LIMIT+1];			/* Nested include file stack	*/
FILE	*files[MAX_FILES+1];				/* Read / write files			*/
char	capture_on = FALSE;
char	filter8 = FALSE;

char	input_buf[BUFFER_SIZE];			/* Default input buffer			*/
char	*line_table[1002], *line_ptr;
int		last_state = 0;						/* Initially there are no states*/
int		call_stack[100], call_level;
int		yyline, pass, last_line;
int		qnx_terminal, has_timeouts;
pid_t	qterm_pid;

struct	state_name	state_table[ MAX_STATE ];
struct	var_entry	*retval, *cmdval, *collect, *modem, *failed, *switcher;
struct	var_entry	*tcap;
struct	var_entry	*lookup_var();
struct	msg			m;

/*--- Command line options ----------*/
char	debug = FALSE;
char	listing = FALSE;
char	verbose = FALSE;
char	echo = TRUE;
char	qterm = FALSE;
char	secure = FALSE;
char	quiet = FALSE;

void
main( int argc, char *argv[] ) {
	int					n, nest;
	struct var_entry	*temp;
	FILE				*fp;
	unsigned			opt;

	if ( argc < 2 ) {
		fprintf( stderr, "%s (%s)\n%s\n%s\n%s\n%s\n",
		VERSION, __DATE__,
#ifdef __USAGE
%C	[options]* filename

options:    -m modem_name                    Modem port to use
            -v variable_name=value           Assign a variable
            -d                               Debugging output
            -e                               No echo
            -l                               Listing only
            -s                               Secure
            -V                               Be verbose
            -q                               Be quiet
            -t                               No timeouts
            -Q                               Started via Qterm
#endif
		"use: qcl [options]* filename",
		"options: -mmodem_name -vvariable_name=value",
		"         -debug -echo -listing_only -secure -Verbose -quiet -Qterm",
		"         -timeouts" );
		exit( -1 );
		}

	signal( SIGINT, &abort_check );

	qterm_pid = getppid();
	qnx_terminal = strncmp( t.name, "qnx", 3 ) == 0;
	modem = lookup_var( "modem$" );			/* Set modem name string */
	_Free( modem->vstr );
	strcpy( modem->vstr = _Malloc( 10 ), "/dev/ser1" );

	opterr = 0;
	while( ( opt = getopt( argc, argv, "m:v:delsVqQt" ) ) != -1 ) {
		switch( opt ) {
			case 'd': 	debug = TRUE;					break;
			case 'e':	echo = FALSE;					break;
			case 'l':	verbose = listing = TRUE;		break;
			case 's':	secure = TRUE;					break;
			case 't':	has_timeouts = FALSE;			break;
			case 'V': 	verbose = TRUE;					break;
			case 'q': 	quiet = TRUE;					break;
			case 'Q':	qterm = TRUE;					break;
			case 'm':	_Free( modem->vstr );
						strcpy( modem->vstr = _Malloc( strlen( optarg )+1 ), optarg );
						break;
			case 'v':	assign( optarg );				break;
			case '?':	fprintf( stderr, "qcl: Unknown option '%s'.\n", optopt );
						exit( EXIT_FAILURE );
			}
		}
qterm = FALSE;

	if ( ( mdm = open( modem->vstr, O_RDWR ) ) == -1 ) {
		fprintf( stderr, "qcl: Unable to open device '%s'.\r\l", modem->vstr );
		exit( EXIT_FAILURE );
		}

	okbd_mode = dev_mode( 0, 0, 0 );

	if ( term_load() == -1 && !quiet ) {
		fprintf( stderr, "qcl: Unable to load tcap entry, forcing quiet mode.\n");
		quiet = TRUE; qterm = FALSE;
		}

	strcpy( input_buf, argv[optind] );
	if ( !strchr( input_buf, '.' ) ) strcat( input_buf, ".qcl" );

	/* Set the modem completely RAW */
	omdm_mode = a_dev_mode( mdm, 0, _DEV_MODES );
	tcflush( mdm, TCIOFLUSH );

	kbd_mode = dev_mode( 0, 0, 0 );
	tcflush( 0, TCIOFLUSH );

	if ( !( nest_fp[ nest = 0 ] = fopen( input_buf, "r" ) ) ) {
		fprintf( stderr, "qcl: Unable to read '%s'.\r\l", input_buf );
		exit( EXIT_FAILURE );
		}

	for ( yyline = 1; yyline <= 1000; ) {		/* Read program into buffer	*/
		if ( !( n = fgetln( nest_fp[ nest ], input_buf, 300 ) ) ) {
			fclose( nest_fp[ nest ] );
			if ( !nest ) break;				/* Done with read, now execute	*/
			--nest;							/* Back out a level of nesting	*/
			continue;						/* Continue reading previous	*/
			}
		if ( input_buf[0] == ';'  ||  input_buf[0] == '"'  ||
			 input_buf[0] == '!'  ||  n <= 1 )
			continue;							/* Skip comment lines		*/
		if ( strncmp( input_buf, "include", 7 ) )
			line_table[ yyline++ ] =  mkstr( input_buf );
		else {
			input_buf[ strlen( input_buf ) - 1 ] = '\0';
			while(isspace(input_buf[strlen(input_buf)-1]))
				input_buf[strlen(input_buf)-1] = '\0';
			if ( !( fp = fopen( input_buf + 8, "r" ) ) ) {
				fprintf( stderr, "qcl: Unable to read '%s'.\r\l", input_buf + 8 );
				exit( EXIT_FAILURE );
				}
			if ( ++nest >= NEST_LIMIT ) {
				fprintf( stderr, "QCL programs too deeply nested.\r\l" );
				exit( EXIT_FAILURE );
				}
			nest_fp[ nest ] = fp;
			}
		}

	line_table[ 1001 ] = line_table[ last_line = yyline ] = "end\n";
	pass = 1;
	line_ptr = line_table[ yyline = 1 ];

	retval = lookup_var( "retval" );
	cmdval = lookup_var( "cmdval" );
	failed = lookup_var( "failed" );
	temp = lookup_var( "user$" );
	if ( temp->vstr ) _Free( temp->vstr );

	{
#include <pwd.h>
	struct	passwd	*PW;
	if ( !( PW = getpwuid( geteuid() ) ) ) 	temp->vstr = "unknown";
	else	strcpy( temp->vstr = _Malloc( strlen( PW->pw_name )+1 ), PW->pw_name );
	}

	temp = lookup_var( "tid" );	temp->vint = getpid();
	temp = lookup_var( "pid" );	temp->vint = getpid();
	temp = lookup_var( "nid" ); temp->vint = getnid();
	temp = lookup_var( "ttynum" );
	strcpy( temp->vstr = _Malloc( strlen( ttyname( 0 ) ) ), ttyname( 0 ) );

	collect = lookup_var( "collect$" );		/* Set collect to be input buffer*/
	if ( collect->vstr ) _Free( collect->vstr );
	collect->vstr = input_buf;
	switcher = collect;						/* Set default for case statements*/

	tcap = lookup_var( "tcap$" );			/* Set terminal name variable	*/
	if ( tcap->vstr ) _Free( tcap->vstr );
	tcap->vstr = t.name;

	files[ 0 ] = stdin;
	for ( n = 1; n > MAX_FILES; ++n )	files[ n ] = NULL;

	yyparse();					/* Parse the program, fixing any ifs		*/

	if ( verbose )
		for ( n = 1; n < 1000; ++n )
			if ( line_table[ n ] )
				fprintf( stderr, "%3d.\t%s", n, line_table[ n ] );

	pass = 2;
	Goto( FALSE );				/* Goto first state in program				*/
	if ( !listing )	yyparse();	/* Execute the program						*/

	done( 0 );
	}

unsigned
a_dev_mode(int fd, unsigned mode, unsigned mask) {
	unsigned rc;
	static int last_fd=-1;

    /* old_mode() will call a_dev_mode, this prevents an infinite loop */
	if (fd==last_fd) return dev_mode(fd,mode,mask);
	
	if ((rc = dev_mode(fd,mode,mask))==-1) {
		perror("dev_mode");
		last_fd = fd;
		if ( mdm != -1 ) tcdropline( mdm, 500 );
		old_mode();		/* may be pointless - depends on fd! */
		exit (EXIT_FAILURE);
		}
	return rc;
	}

#include <term.h>
#define ct __cur_term
#define t term_state
void
done( int status ) {
	if ( qnx_terminal ) write( 1, "\033R", 2 );
//	if ( !quiet && ct->_cc != -1 )
//		console_read( ct->_cc, 0, 0, t.scrbuf, t.screen_amount, &t.row, &t.col, NULL );
	exit( status );
	}

char *
mkstr( char *str ) {
	char *p;
	int i;

	i = strlen( str ) + 1;
	memcpy( p = _Malloc( i ), str, i );
	return p;
	}

char *
_Malloc( int amount ) {
	char *p;

	if ( !( p = malloc( amount ) ) ) {
#include <sys/osinfo.h>
		struct _osinfo os;
		if ( qnx_osinfo( getnid(), &os ) != -1 )
			fprintf( stderr, "\r\lQCL: System memory free: %uk of %uk\r\l", os.freememk, os.totmemk );
		fprintf( stderr, "\r\lQCL: Out of memory trying to malloc %d bytes (1)\r\l", amount );
		exit( EXIT_FAILURE );
		}
	return p;
	}

void
_Free( char *p ) {
	free( p );
	}

/*----------------------------------------------------------------------------
IMPORT:	nchars - The number of characters to read.
			If 0, return the number of characters waiting
			If positive, get the requested number of characters
			If -1, Get a '\r' terminated line of text up to 80 characters.
		timeout - tenths of second to wait
			If -1, don't wait
		buf - place to put received data
EXPORT: Number of characters read or waiting to be read.
----------------------------------------------------------------------------*/
read_mdm( int nchars, int timeout, uchar *buf ) {
	int n = 0, i;

	if ( !nchars )	return dev_ischars( mdm );
	if ( timeout == -1 ) timeout = 0;

	while( nchars && (i = dev_read( mdm, buf, nchars, nchars, 0, timeout, 0, 0 )) ) {
		if ( i == -1 ) perror( "read_mdm()" );
		buf += i;
		nchars -= i;
		n += i;
		}
	/* @@@ Do null and 8th bit stripping */
	if ( capture_fd != -1  &&  capture_on ) write( capture_fd, buf, n );
	return n;
	}

/*----------------------------------------------------------------------------
IMPORT:	nchars - Number of characters to write
			If 0, the length of the buffer is written
		echo - Flag indicating if an echo should come back for each character
----------------------------------------------------------------------------*/
void
write_mdm( unsigned nchars, unsigned mdm_echo, uchar *buf ) {
	unsigned	i, n;
	unsigned	char *p, c;

	if ( !nchars )	nchars = strlen( buf );

	if ( mdm_echo == 2 )
		write( mdm, buf, nchars );
	else
		for ( p = buf; p < buf + nchars; ++p ) {
			write( mdm, p, 1 );
			if ( mdm_echo ) {
				for ( i = 2; i; --i ) {
					n = read_mdm( 1, 5, &c );
					if ( n==1  &&  c == ( *p & (filter8 ? 0x7f : 0xff)) ) break;
					}
				}
			else
				delay( 50 );
			}

	if ( echo )	{
		qterm_fput( buf, nchars );
		}
	}

strip_nulls( char *buf, int nchars ) {
	char	*p1, *p2;
	int		count = 0;

	for( p1 = p2 = buf; p1 < buf + nchars; ++p1 )
		if ( *p1 ) {
			*p2++ = *p1 & ( filter8 ? 0x7f : 0xff );
			count++;
			}
	*p2 = 0;			/* End the buffer with a null for string handling	*/
	return count;
	}

void
assign( char *name ) {
	struct	var_entry *v;
	char	*value;

	if ( !( value = strchr( name, '=' ) ) ) {
		fprintf( stderr, "qcl: Variable assignment syntax error.\r\l" );
		exit( EXIT_FAILURE );
		}
	*value = '\0';							/* Replace '=' with a null		*/
	++value;
	if ( lookup_kword( name ) ) {
		fprintf( stderr, "qcl: Cannot assign a value to a keyword.\n" );
		exit( EXIT_FAILURE );
		}
	v = (struct var_entry *) lookup_var( name );
	if ( v->vstr ) {
#ifdef DEBUG
fprintf( stderr, "In assign " );
#endif
		_Free( v->vstr );
		v->vstr = mkstr( value );
		}
	else	v->vint = atoi( value );
	}

void
yyerror( char *fmt, ... ) {
	char msgbuf[ 80 ];
	int n, low;
	va_list arglist;

	va_start( arglist, fmt );
	vsprintf( msgbuf, fmt, arglist );
	va_end( arglist );
	fprintf( stderr, "\r\lLine %d: %s", yyline, msgbuf );
	fprintf( stderr, "\r\l" );
	low = yyline;
	if ( low > 1 ) --low;
	for ( n = low; n < yyline + 3; ++n )
		if ( line_table[ n ] )
			fprintf( stderr, "%3d.\t%s\r\l", n, line_table[ n ] );
	}

fgetln( FILE *fp, char *line, int maxlen ) {
	int len = 0;
	char *p;

	line[0] = '\0';
	if ( p = fgets( line, maxlen, fp ) )
		len = strlen( line );
	return len;
	}

qterm_fput( char *buf, int count ) {
	if ( qterm ) {
		m.row = m.col = -1;
		m.len = count;
		if ( Send( qterm_pid, &m, &m, sizeof( struct msg ), sizeof( struct msg ) ) == -1 ) {
			qterm = FALSE;
			return Display( -1, buf, 0 ) == count ? 0 : -1;
			}
		}
	else
		return Display( -1, buf, 0 ) == count ? 0 : -1;
	return 0; // should never get here -- just to shutup compiler
	}
