/*				29-Mar-90  4:49:56pm										*/

/*--------------------------------------------------------------------------*/
/*  History: yylex.c, V0.0, ??-Nov-86 11:24:14am, Dan Dodge,	Baseline	*/
/*  				  V0.1, 13-Jan-87 11:24:14am, Dan Hildebrand			*/
/*----------------------------------------------------------------------------

----------------------------------------------------------------------------*/
#include "qcl.h"
#include "y.tab.h"

extern YYSTYPE yylval;
extern int yyline, pass;
extern char *line_table[], *line_ptr;
extern struct state_name state_table[ MAX_STATE ];
extern char debug;

struct var_entry *lookup_var();

char	name[ NAME_LENGTH+1 ];	/* Buffer to build variable names			*/
char	sval[ 257 ];			/* Buffer to build strings					*/
struct	var_entry *var_list;
char	*line_ptr_save;			/* Needed because of YACC's look ahead		*/

#define	MAX_SCRATCH		200
char	scratch[ MAX_SCRATCH ];	/* Space for yylex to place extra strings	*/
char	*scr_ptr = scratch;		/* Pointer to free space in scratch[]		*/

#define	BOTH	0
#define	PASS1	1
#define	PASS2	2

/* Functions must be both passes, commands must be pass2	*/
/* Conditionals and high level structures must be both		*/
struct kword_entry {
		char	*kname;					/* Ascii representation				*/
		int		kvalue2;				/* Keyword value on pass2			*/
		int		kvalue1;				/* Keyword value on pass1			*/
		} kword_table[] ={
			"access",		_ACCESS,		_ACCESS,
			"beep",			_BEEP,			_COMMENT,
			"box",			_BOX,			_COMMENT,
			"break",		_BREAK,			_COMMENT,
			"breakable",	_BREAKABLE,		_COMMENT,
			"call",			_CALL,			_COMMENT,
			"carrier",		_CARRIER,		_CARRIER,
			"capture",		_CAPTURE,		_COMMENT,
			"case",			_CASE,			_CASE,
			"cd",			_CD,			_COMMENT,
			"clear",		_CLEAR,			_COMMENT,
			"close",		_CLOSE,			_COMMENT,
			"collect",		_COLLECT,		_COMMENT,
			"colour",		_COLOUR,		_COMMENT,
			"color",		_COLOUR,		_COMMENT,
			"comment",		_COMMENT,		_COMMENT,
			"cursor",		_CURSOR,		_COMMENT,
			"date$",		_DATE,			_DATE,
			"dec",			_DEC,			_COMMENT,
			"display",		_DISPLAY,		_COMMENT,
			"drop",			_DROP,			_DROP,
			"dropline",		_DROPLINE,		_COMMENT,
			"dump",			_DUMP,			_COMMENT,
			"echo",			_ECHO,			_COMMENT,
			"else",			_ELSE,			_ELSE,
			"emit", 		_EMIT,			_COMMENT,
			"end",			_END,			_END,
			"eof",			_EOFILE,		_EOFILE,
			"equal",		_FEQUAL,		_FEQUAL,
			"exit",			_EXIT,			_COMMENT,
			"fast",			_FAST,			_COMMENT,
			"filter8",		_FILTER8,		_COMMENT,
			"flush",		_FLUSH,			_COMMENT,
			"get",			_GET,			_GET,
			"goto",			_GOTO,			_COMMENT,
			"if",			_IF,			_IF,
			"input",		_INPUT,			_INPUT,
			"itos",			_ITOS,			_ITOS,
			"jump",			_JUMP,			_COMMENT,
			"length",		_LENGTH,		_LENGTH,
			"match",		_MATCH,			_MATCH,
			"off",			_OFF,			_COMMENT,
			"on",			_ON,			_COMMENT,
			"open",			_OPEN,			_COMMENT,
			"proto_emit",	_PROTO,			_COMMENT,
			"read",			_READQ,			_READQ,
			"repeat",		_REPEAT,		_REPEAT,
			"return",		_RETURN,		_RETURN,
			"secure",		_SECURE,		_COMMENT,
			"shell",		_SHELL,			_COMMENT,
			"sleep",		_SLEEP,			_COMMENT,
			"state",		_STATE,			_STATE,
			"stoi",			_STOI,			_STOI,
			"switch",		_SWITCH,		_COMMENT,
			"take",			_TAKE,			_TAKE,
			"termload",		_TERMLOAD,		_COMMENT,
			"time$",		_TIME,			_TIME,
			"tolower",		_TOLOWER,		_TOLOWER,
			"toupper",		_TOUPPER,		_TOUPPER,
			"trace",		_TRACE,			_COMMENT,
			"unbreakable",	_UNBREAKABLE,	_COMMENT,
			"waitfor",		_WAITFOR,		_WAITFOR,
			"while",		_WHILE,			_WHILE,
			"write",		_WRITEQ,		_COMMENT,
			NULL
			};

yylex() {
	int		token, c;

start:
	while( *line_ptr == '\t'  ||  *line_ptr == ' ' )	++line_ptr;

	if ( !line_ptr  ||  !( c = *line_ptr++ ) ) {
		line_ptr = line_table[ ++yyline ];
		if ( pass == 2  &&  debug )
			fprintf( stderr, "EXEC %d. %s", yyline, line_ptr );
		goto start;
		}

	line_ptr_save = line_ptr - 1;

	if ( c == '"' ) {
		yylval.cpval = getstring( '"' );
		token = _STRING;
		}
	else if ( isalpha( c )  ||  c == '_' ) {
		if ( !( token = lookup_kword( getname( c ) ) ) ) {
			yylval.vpval = lookup_var( name );
			/* Know if string or int based on '$' as last character			*/
			token = yylval.vpval->vstr ? _SVAR : _VAR; /* Return str or int	*/
			}
		else if ( token == _COMMENT ) {
			line_ptr = "";				/* Will fetch new line on next call */
			token = _EOL;
			}
		}
	else if ( isdigit( c ) ) {
		yylval.ival = getcon( c );
		token = _CON;
		}
	else if ( c == '-'  &&  *line_ptr == '-' ) {
		++line_ptr;
		token = _DEC;
		}
	else if ( c == '+'  &&  *line_ptr == '+' ) {
		++line_ptr;
		token = _INC;
		}
	else if ( c == '='  &&  *line_ptr == '=' ) {
		++line_ptr;
		yylval.ival = '==';
		token = _COMP;
		}
	else if ( c == '!'  &&  *line_ptr == '=' ) {
		++line_ptr;
		yylval.ival = '!=';
		token = _COMP;
		}
	else if ( c == '<'  &&  *line_ptr == '=' ) {
		yylval.ival = '<=';
		++line_ptr;
		token = _COMP;
		}
	else if ( c == '='  &&  *line_ptr == '<' ) {
		yylval.ival = '<=';
		++line_ptr;
		token = _COMP;
		}
	else if ( c == '>'  &&  *line_ptr == '=' ) {
		yylval.ival = '>=';
		++line_ptr;
		token = _COMP;
		}
	else if ( c == '='  &&  *line_ptr == '>' ) {
		yylval.ival = '>=';
		++line_ptr;
		token = _COMP;
		}
	else if ( c == '<' ) {
		yylval.ival = '<';
		token = _COMP;
		}
	else if ( c == '>' ) {
		yylval.ival = '>';
		token = _COMP;
		}
	else if ( c == '|'  &&  *line_ptr == '|' ) {
		++line_ptr;
		token = _OR;
		}
	else if ( c == '&'  &&  *line_ptr == '&' ) {
		++line_ptr;
		token = _AND;
		}
	else if ( c == '\n' ) 	token = _EOL;
	else if ( c == '*' ) {
		yylval.ival = '*';
		token = _PUTE;
		}
	else if ( c == '/' ) {
		yylval.ival = '/';
		token = _PUTE;
		}
	else if ( c == '%' ) {
		yylval.ival = '%';
		token = _PUTE;
		}
	else if ( c == ';' ) {
		line_ptr = "";
		token = _EOL;
		}
	else	token = c;

	return( token );
	}


/*
 * Read in an identifier.
 * 'c' is the first character of the name.
 */
char *
getname( char c ) {
	char *np, *last;

	np = name;
	last = &name[ NAME_LENGTH ];

	do {
		if( np < last )	*np++ = c;
		c = *line_ptr++;
		}
	while( isalnum( c )  ||  c == '_'  ||  c == '$' );

	--line_ptr;
	*np = '\0';
	return( name );
	}


/*
 * Read in a number.
 * 'c' is the first character of the name.
 */
getcon( char c ) {
	char *np, *last;
	int val = 0;

	np = name;
	last = &name[ NAME_LENGTH ];

	if ( c == '0'  &&  *line_ptr == 'x' ) {		/* Hex constants		*/
		++line_ptr;
		do {
			if( np < last )	*np++ = c;
			c = *line_ptr++;
			}
		while( isxdigit( c ) );
		*np = '\0';
		val = atoh( name );
		}
	else {										/* Decimal constants	*/
		do {
			if( np < last )	*np++ = c;
			c = *line_ptr++;
			}
		while( isdigit( c ) );
		*np = '\0';
		val = atoi( name );
		}

	--line_ptr;
	return( val );
	}


/*
 * Read in a string.
 * 'term' is the string terminator.
 */
char *
getstring( char term ) {
	char *np, *last, c;

	np = sval;
	last = &sval[ 256 ];

	while( *line_ptr && (c = *line_ptr++) != term && c != '\n'  &&  np < last )
		*np++ = ( c == '\\' ) ? map_escape( *line_ptr++ ) : c;

	if ( c == '\n' )	--line_ptr;
	*np = '\0';
	strcpy( np = scr_ptr, sval );	/* Save string in the scratch buffer	*/
	scr_ptr += strlen( sval ) + 1;	/* Point to next part of free space		*/
	if ( scr_ptr > scratch + MAX_SCRATCH ) {
		fprintf( stderr, "qcl: Scratch overflow\r\l" );
		exit( EXIT_FAILURE );
		}
	return( np );					/* Return pointer to string in scratch[]*/
	}

map_escape( c )
	char c;
	{
	char buf[ 3 ];

	switch( c ) {
		case 'b': 	return( '\b' );			/* Backspace		*/
		case 'f': 	return( '\f' );			/* Formfeed			*/
		case 'l': 	return( '\l' );			/* Linefeed			*/
		/* @@@ Need \n to \r\l expansion						*/
		case 'n': 	return( '\n' );			/* Newline			*/
		case 'r': 	return( '\r' );			/* Carriage return	*/
		case 't': 	return( '\t' );			/* Tab				*/

		case 'x': 	buf[ 0 ] = *line_ptr++;	/* Hex constant		*/
					buf[ 1 ] = *line_ptr++;
					buf[ 2 ] = 0;
					return( atoh( buf ) );
		}
	return( c );
	}

/* Return the token number associated with a string	*/
lookup_kword( str )
	char *str;
	{
	struct kword_entry  *kp;

	/* Pre-adjust the search start point	*/
	kp = ( str[0] >= 'i' ) ? &kword_table[ 30 ] : &kword_table[ 0 ];

	for ( ; kp->kname; ++kp )
		if ( kp->kname[0] == str[0]  &&  !strcmp( kp->kname, str ) )
			return( pass == 1 ? kp->kvalue1 : kp->kvalue2 );
/*
	for ( kp = kword_table; kp->kname; ++kp )
		if ( !strcmp( kp->kname, str ) )
			return( ( pass == 1 &&  kp->pass == PASS2 )	? _COMMENT
														: kp->kvalue );
*/
	return( 0 );		/* Not a keyword		*/
	}

/*
Example:
    return bsearch( kword_table,
					kword_table + (sizeof(kword_table) / sizeof(kword_entry)),
					nam );

struct kword_entry
*bsearch( lo1, hi, nam )
	struct kword_entry *lo1, *hi;
	char *nam;
	{
	int i;
	struct kword_entry *chk, *lo = lo1;

	for (;;) {
		chk = lo + ( ( hi - lo ) >> 1);
		if ( !(i = strcmp( chk->kname, nam ) ) ) return( chk );
		if ( chk == lo ) return( NULL );
		if ( i < 0 ) lo = chk;
		else hi = chk;
		}
	}
*/

struct var_entry *
lookup_var( char *str ) {
	struct var_entry *vp, *new;
	int len;

	/*--- Attempt to lookup variable before allocating space for it.		*/
	for( vp = var_list; vp; vp = vp->vlink ) {
		if ( *str == *vp->vname  &&  strcmp( str, vp->vname ) == 0 )
			return( vp );
		if ( vp->vlink == NULL )
			break;
		}

	if( !( new = calloc( sizeof( struct var_entry ) +
						 ( len = strlen( str ) ) + 1, 1 ) ) ) {
		yyerror( "Out of memory allocating variable %s", str );
		done( -1 );
		}
	if ( !var_list ) var_list = new;
	else			 vp->vlink = new;

	memcpy( new->vname, str, len + 1 );
	if ( str[ len-1 ] == '$' ) /* Vars ending in '$' are made into strings	*/
		new->vstr = calloc( 1, 1 );
	return( new );
	}

void
dump_variables() {
	struct var_entry *vp;

	printf( "Variable dump:\r\l" );
	for( vp = var_list; vp; vp = vp->vlink ) {
		printf( "%s: ", vp->vname );
		if ( vp->vstr ) printf( "|%s|\r\l", vp->vstr );
		else			printf( "%d\r\l", vp->vint );
		}
	printf( "--------------\r\l" );
	}

struct var_entry *
make_unique_name() {
	static int num = 0;
	char buf[ 7 ];

	sprintf( buf, "TMP%d", ++num );

	return( lookup_var( buf ) );
	}

void
reclaim_scratch() {
	scr_ptr = scratch;			/* Reclaim scratch string space		*/
	}
